
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kommons.math

import ru.dbotthepony.kommons.util.IStruct2d
import ru.dbotthepony.kommons.util.IStruct2f
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.util.IStruct2l

private inline fun span(x0: Float, x1: Float): Float {
	return x1 - x0
}

private inline fun span(x0: Double, x1: Double): Double {
	return x1 - x0
}

/**
 * Whenever two segments on one axis intersect
 */
fun intersectSegments(a0: Float, a1: Float, b0: Float, b1: Float): Boolean {
	return a0 in b0 .. b1 || a1 in b0 .. b1 || b0 in a0 .. a1 || b1 in a0 .. a1
}

/**
 * Whenever two segments on one axis intersect
 */
fun intersectSegments(a0: Double, a1: Double, b0: Double, b1: Double): Boolean {
	return a0 in b0 .. b1 || a1 in b0 .. b1 || b0 in a0 .. a1 || b1 in a0 .. a1
}

/**
 * Whenever two segments on one axis intersect
 */
fun intersectSegments(a0: Int, a1: Int, b0: Int, b1: Int): Boolean {
	return a0 in b0 .. b1 || a1 in b0 .. b1 || b0 in a0 .. a1 || b1 in a0 .. a1
}

/**
 * Whenever two segments on one axis intersect
 */
fun intersectSegments(a0: Long, a1: Long, b0: Long, b1: Long): Boolean {
	return a0 in b0 .. b1 || a1 in b0 .. b1 || b0 in a0 .. a1 || b1 in a0 .. a1
}

fun intersectRectangles(
	thisMinsX: Float,
	thisMinsY: Float,
	thisMaxsX: Float,
	thisMaxsY: Float,

	otherMinsX: Float,
	otherMinsY: Float,
	otherMaxsX: Float,
	otherMaxsY: Float,
): Boolean {
	return intersectSegments(thisMinsX, thisMaxsX, otherMinsX, otherMaxsX) && intersectSegments(thisMinsY, thisMaxsY, otherMinsY, otherMaxsY)
}

fun rectangleContainsRectangle(
	thisMinsX: Float,
	thisMinsY: Float,
	thisMaxsX: Float,
	thisMaxsY: Float,

	otherMinsX: Float,
	otherMinsY: Float,
	otherMaxsX: Float,
	otherMaxsY: Float,
): Boolean {
	if (span(thisMinsX, thisMaxsX) < span(otherMinsX, otherMaxsX) || span(thisMinsY, thisMaxsY) < span(otherMinsY, otherMaxsY))
		return false

	return  otherMinsX in thisMinsX .. thisMaxsX &&
			otherMaxsX in thisMinsX .. thisMaxsX &&
			otherMinsY in thisMinsY .. thisMaxsY &&
			otherMaxsY in thisMinsY .. thisMaxsY
}

fun intersectRectangles(
	thisMins: IStruct2f,
	thisMaxs: IStruct2f,

	otherMins: IStruct2f,
	otherMaxs: IStruct2f
): Boolean {
	val (thisMinsX, thisMinsY) = thisMins
	val (thisMaxsX, thisMaxsY) = thisMaxs

	val (otherMinsX, otherMinsY) = otherMins
	val (otherMaxsX, otherMaxsY) = otherMaxs

	return intersectRectangles(thisMinsX, thisMinsY, thisMaxsX, thisMaxsY, otherMinsX, otherMinsY, otherMaxsX, otherMaxsY)
}

fun rectangleContainsRectangle(
	thisMins: IStruct2f,
	thisMaxs: IStruct2f,

	otherMins: IStruct2f,
	otherMaxs: IStruct2f
): Boolean {
	val (thisMinsX, thisMinsY) = thisMins
	val (thisMaxsX, thisMaxsY) = thisMaxs

	val (otherMinsX, otherMinsY) = otherMins
	val (otherMaxsX, otherMaxsY) = otherMaxs

	return rectangleContainsRectangle(thisMinsX, thisMinsY, thisMaxsX, thisMaxsY, otherMinsX, otherMinsY, otherMaxsX, otherMaxsY)
}

fun intersectRectangles(
	thisMinsX: Double,
	thisMinsY: Double,
	thisMaxsX: Double,
	thisMaxsY: Double,

	otherMinsX: Double,
	otherMinsY: Double,
	otherMaxsX: Double,
	otherMaxsY: Double,
): Boolean {
	return intersectSegments(thisMinsX, thisMaxsX, otherMinsX, otherMaxsX) && intersectSegments(thisMinsY, thisMaxsY, otherMinsY, otherMaxsY)
}

fun rectangleContainsRectangle(
	thisMinsX: Double,
	thisMinsY: Double,
	thisMaxsX: Double,
	thisMaxsY: Double,

	otherMinsX: Double,
	otherMinsY: Double,
	otherMaxsX: Double,
	otherMaxsY: Double,
): Boolean {
	if (span(thisMinsX, thisMaxsX) < span(otherMinsX, otherMaxsX) || span(thisMinsY, thisMaxsY) < span(otherMinsY, otherMaxsY))
		return false

	return  otherMinsX in thisMinsX .. thisMaxsX &&
			otherMaxsX in thisMinsX .. thisMaxsX &&
			otherMinsY in thisMinsY .. thisMaxsY &&
			otherMaxsY in thisMinsY .. thisMaxsY
}

fun intersectRectangles(
	thisMins: IStruct2d,
	thisMaxs: IStruct2d,

	otherMins: IStruct2d,
	otherMaxs: IStruct2d
): Boolean {
	val (thisMinsX, thisMinsY) = thisMins
	val (thisMaxsX, thisMaxsY) = thisMaxs

	val (otherMinsX, otherMinsY) = otherMins
	val (otherMaxsX, otherMaxsY) = otherMaxs

	return intersectRectangles(thisMinsX, thisMinsY, thisMaxsX, thisMaxsY, otherMinsX, otherMinsY, otherMaxsX, otherMaxsY)
}

fun rectangleContainsRectangle(
	thisMins: IStruct2d,
	thisMaxs: IStruct2d,

	otherMins: IStruct2d,
	otherMaxs: IStruct2d
): Boolean {
	val (thisMinsX, thisMinsY) = thisMins
	val (thisMaxsX, thisMaxsY) = thisMaxs

	val (otherMinsX, otherMinsY) = otherMins
	val (otherMaxsX, otherMaxsY) = otherMaxs

	return rectangleContainsRectangle(thisMinsX, thisMinsY, thisMaxsX, thisMaxsY, otherMinsX, otherMinsY, otherMaxsX, otherMaxsY)
}

private inline fun span(x0: Int, x1: Int): Int {
	return x1 - x0
}

fun intersectRectangles(
	thisMinsX: Int,
	thisMinsY: Int,
	thisMaxsX: Int,
	thisMaxsY: Int,

	otherMinsX: Int,
	otherMinsY: Int,
	otherMaxsX: Int,
	otherMaxsY: Int,
): Boolean {
	return intersectSegments(thisMinsX, thisMaxsX, otherMinsX, otherMaxsX) && intersectSegments(thisMinsY, thisMaxsY, otherMinsY, otherMaxsY)
}

fun rectangleContainsRectangle(
	thisMinsX: Int,
	thisMinsY: Int,
	thisMaxsX: Int,
	thisMaxsY: Int,

	otherMinsX: Int,
	otherMinsY: Int,
	otherMaxsX: Int,
	otherMaxsY: Int,
): Boolean {
	if (span(thisMinsX, thisMaxsX) < span(otherMinsX, otherMaxsX) || span(thisMinsY, thisMaxsY) < span(otherMinsY, otherMaxsY))
		return false

	return  otherMinsX in thisMinsX .. thisMaxsX &&
			otherMaxsX in thisMinsX .. thisMaxsX &&
			otherMinsY in thisMinsY .. thisMaxsY &&
			otherMaxsY in thisMinsY .. thisMaxsY
}

fun intersectRectangles(
	thisMins: IStruct2i,
	thisMaxs: IStruct2i,

	otherMins: IStruct2i,
	otherMaxs: IStruct2i
): Boolean {
	val (thisMinsX, thisMinsY) = thisMins
	val (thisMaxsX, thisMaxsY) = thisMaxs

	val (otherMinsX, otherMinsY) = otherMins
	val (otherMaxsX, otherMaxsY) = otherMaxs

	return intersectRectangles(thisMinsX, thisMinsY, thisMaxsX, thisMaxsY, otherMinsX, otherMinsY, otherMaxsX, otherMaxsY)
}

fun rectangleContainsRectangle(
	thisMins: IStruct2i,
	thisMaxs: IStruct2i,

	otherMins: IStruct2i,
	otherMaxs: IStruct2i
): Boolean {
	val (thisMinsX, thisMinsY) = thisMins
	val (thisMaxsX, thisMaxsY) = thisMaxs

	val (otherMinsX, otherMinsY) = otherMins
	val (otherMaxsX, otherMaxsY) = otherMaxs

	return rectangleContainsRectangle(thisMinsX, thisMinsY, thisMaxsX, thisMaxsY, otherMinsX, otherMinsY, otherMaxsX, otherMaxsY)
}

private inline fun span(x0: Long, x1: Long): Long {
	return x1 - x0
}

fun intersectRectangles(
	thisMinsX: Long,
	thisMinsY: Long,
	thisMaxsX: Long,
	thisMaxsY: Long,

	otherMinsX: Long,
	otherMinsY: Long,
	otherMaxsX: Long,
	otherMaxsY: Long,
): Boolean {
	return intersectSegments(thisMinsX, thisMaxsX, otherMinsX, otherMaxsX) && intersectSegments(thisMinsY, thisMaxsY, otherMinsY, otherMaxsY)
}

fun rectangleContainsRectangle(
	thisMinsX: Long,
	thisMinsY: Long,
	thisMaxsX: Long,
	thisMaxsY: Long,

	otherMinsX: Long,
	otherMinsY: Long,
	otherMaxsX: Long,
	otherMaxsY: Long,
): Boolean {
	if (span(thisMinsX, thisMaxsX) < span(otherMinsX, otherMaxsX) || span(thisMinsY, thisMaxsY) < span(otherMinsY, otherMaxsY))
		return false

	return  otherMinsX in thisMinsX .. thisMaxsX &&
			otherMaxsX in thisMinsX .. thisMaxsX &&
			otherMinsY in thisMinsY .. thisMaxsY &&
			otherMaxsY in thisMinsY .. thisMaxsY
}

fun intersectRectangles(
	thisMins: IStruct2l,
	thisMaxs: IStruct2l,

	otherMins: IStruct2l,
	otherMaxs: IStruct2l
): Boolean {
	val (thisMinsX, thisMinsY) = thisMins
	val (thisMaxsX, thisMaxsY) = thisMaxs

	val (otherMinsX, otherMinsY) = otherMins
	val (otherMaxsX, otherMaxsY) = otherMaxs

	return intersectRectangles(thisMinsX, thisMinsY, thisMaxsX, thisMaxsY, otherMinsX, otherMinsY, otherMaxsX, otherMaxsY)
}

fun rectangleContainsRectangle(
	thisMins: IStruct2l,
	thisMaxs: IStruct2l,

	otherMins: IStruct2l,
	otherMaxs: IStruct2l
): Boolean {
	val (thisMinsX, thisMinsY) = thisMins
	val (thisMaxsX, thisMaxsY) = thisMaxs

	val (otherMinsX, otherMinsY) = otherMins
	val (otherMaxsX, otherMaxsY) = otherMaxs

	return rectangleContainsRectangle(thisMinsX, thisMinsY, thisMaxsX, thisMaxsY, otherMinsX, otherMinsY, otherMaxsX, otherMaxsY)
}

fun intersectCirclePoint(
	circleX: Float,
	circleY: Float,
	circleRadius: Float,

	x: Float,
	y: Float
): Boolean {
	return ((circleX - x) * (circleX - x) + (circleY - y) * (circleY - y)) <= circleRadius * circleRadius
}

fun intersectCirclePoint(
	circleX: Double,
	circleY: Double,
	circleRadius: Double,

	x: Double,
	y: Double
): Boolean {
	return ((circleX - x) * (circleX - x) + (circleY - y) * (circleY - y)) <= circleRadius * circleRadius
}

fun intersectCircleRectangle(
	circleX: Float,
	circleY: Float,
	circleRadius: Float,

	rectMinsX: Float,
	rectMinsY: Float,
	rectMaxsX: Float,
	rectMaxsY: Float,
): Boolean {
	// test if circle's center is inside rectangle
	if (circleX in rectMinsX .. rectMaxsX && circleY in rectMinsY .. rectMaxsY) {
		return true
	}

	// test intersection between rectangle and circle's outer square
	// if they don't intersect there is no way circle can be intersecting with this rect
	if (!intersectRectangles(
		circleX - circleRadius,
		circleY - circleRadius,
		circleX + circleRadius,
		circleY + circleRadius,
		rectMinsX,
		rectMinsY,
		rectMaxsX,
		rectMaxsY,
	)) {
		return false
	}

	return intersectCirclePoint(circleX, circleY, circleRadius, rectMinsX, rectMinsY) ||
			intersectCirclePoint(circleX, circleY, circleRadius, rectMinsX, rectMaxsY) ||
			intersectCirclePoint(circleX, circleY, circleRadius, rectMaxsX, rectMaxsY) ||
			intersectCirclePoint(circleX, circleY, circleRadius, rectMaxsX, rectMinsY)
}

fun intersectCircleRectangle(
	circleX: Float,
	circleY: Float,
	circleRadius: Float,

	mins: IStruct2f,
	maxs: IStruct2f,
): Boolean {
	val (rectMinsX, rectMinsY) = mins
	val (rectMaxsX, rectMaxsY) = maxs

	return intersectCircleRectangle(circleX, circleY, circleRadius, rectMinsX, rectMinsY, rectMaxsX, rectMaxsY)
}

fun intersectCircleRectangle(
	circlePos: IStruct2f,
	circleRadius: Float,

	mins: IStruct2f,
	maxs: IStruct2f,
): Boolean {
	val (circleX, circleY) = circlePos
	val (rectMinsX, rectMinsY) = mins
	val (rectMaxsX, rectMaxsY) = maxs

	return intersectCircleRectangle(circleX, circleY, circleRadius, rectMinsX, rectMinsY, rectMaxsX, rectMaxsY)
}

fun intersectCircleRectangle(
	circlePos: IStruct2f,
	circleRadius: Float,

	rectMinsX: Float,
	rectMinsY: Float,
	rectMaxsX: Float,
	rectMaxsY: Float,
): Boolean {
	val (circleX, circleY) = circlePos

	return intersectCircleRectangle(circleX, circleY, circleRadius, rectMinsX, rectMinsY, rectMaxsX, rectMaxsY)
}

fun intersectCircleRectangle(
	circleX: Double,
	circleY: Double,
	circleRadius: Double,

	rectMinsX: Double,
	rectMinsY: Double,
	rectMaxsX: Double,
	rectMaxsY: Double,
): Boolean {
	// test if circle's center is inside rectangle
	if (circleX in rectMinsX .. rectMaxsX && circleY in rectMinsY .. rectMaxsY) {
		return true
	}

	// test intersection between rectangle and circle's outer square
	// if they don't intersect there is no way circle can be intersecting with this rect
	if (!intersectRectangles(
		circleX - circleRadius,
		circleY - circleRadius,
		circleX + circleRadius,
		circleY + circleRadius,
		rectMinsX,
		rectMinsY,
		rectMaxsX,
		rectMaxsY,
	)) {
		return false
	}

	return intersectCirclePoint(circleX, circleY, circleRadius, rectMinsX, rectMinsY) ||
			intersectCirclePoint(circleX, circleY, circleRadius, rectMinsX, rectMaxsY) ||
			intersectCirclePoint(circleX, circleY, circleRadius, rectMaxsX, rectMaxsY) ||
			intersectCirclePoint(circleX, circleY, circleRadius, rectMaxsX, rectMinsY)
}

fun intersectCircleRectangle(
	circleX: Double,
	circleY: Double,
	circleRadius: Double,

	mins: IStruct2d,
	maxs: IStruct2d,
): Boolean {
	val (rectMinsX, rectMinsY) = mins
	val (rectMaxsX, rectMaxsY) = maxs

	return intersectCircleRectangle(circleX, circleY, circleRadius, rectMinsX, rectMinsY, rectMaxsX, rectMaxsY)
}

fun intersectCircleRectangle(
	circlePos: IStruct2d,
	circleRadius: Double,

	mins: IStruct2d,
	maxs: IStruct2d,
): Boolean {
	val (circleX, circleY) = circlePos
	val (rectMinsX, rectMinsY) = mins
	val (rectMaxsX, rectMaxsY) = maxs

	return intersectCircleRectangle(circleX, circleY, circleRadius, rectMinsX, rectMinsY, rectMaxsX, rectMaxsY)
}

fun intersectCircleRectangle(
	circlePos: IStruct2d,
	circleRadius: Double,

	rectMinsX: Double,
	rectMinsY: Double,
	rectMaxsX: Double,
	rectMaxsY: Double,
): Boolean {
	val (circleX, circleY) = circlePos

	return intersectCircleRectangle(circleX, circleY, circleRadius, rectMinsX, rectMinsY, rectMaxsX, rectMaxsY)
}
