package ru.dbotthepony.kommons.random

import java.util.random.RandomGenerator

/**
 * From [gjrand website](https://gjrand.sourceforge.net):
 *
 * The gjrand project is about pseudo-random number generation for the purpose of simulations,
 * Monte-Carlo integration, computer games and the like. It is intended that statistical
 * properties be extremely good for such purposes. However, gjrand is not suitable as a
 * high security random number generator for cryptography, the state lottery draw, etc.
 */
open class GJRAND64Random protected constructor(
	@JvmField
	protected var s0: Long,
	@JvmField
	protected var s1: Long,
	@JvmField
	protected var s2: Long,
	@JvmField
	protected var s3: Long,
	marker: Nothing?
) : RandomGenerator {
	constructor(seed0: Long, seed1: Long) : this(seed0, seed1, SEED2_INIT, SEED3_INIT, null) {
		for (i in 0 until 14)
			nextLong()
	}

	constructor(seed: Long) : this(seed, 0L)

	protected fun reinitialize(seed: Long) {
		reinitialize(seed, 0L)
	}

	protected fun reinitialize(seed0: Long, seed1: Long) {
		s0 = seed0
		s1 = seed1
		s2 = SEED2_INIT
		s3 = SEED3_INIT

		for (i in 0 until 14)
			nextLong()
	}

	final override fun nextLong(): Long {
		s1 += s2
		s0 = s0.rotateLeft(32)
		s2 = s2 xor s1
		s3 += 0x55aa96a5L
		s0 += s1
		s2 = s2.rotateLeft(23)
		s1 = s1 xor s0
		s0 += s2
		s1 = s1.rotateLeft(19)
		s2 += s0
		s1 += s3
		return s0
	}

	companion object {
		const val SEED2_INIT = 2000001L
		const val SEED3_INIT = 0L
	}
}
